require 'custom_attributes/search_query'

module CustomAttributes
  class FluentSearchQuery
    def initialize(customizable)
      @search_query = CustomAttributes::SearchQuery.new
      @search_query.customizable = customizable unless customizable.nil?

      @raw_results = nil
      @field_list = {}
      @field_queries = {}
      @field_config = {}
    end

    def query(query)
      @search_query.query = query

      self
    end

    def field_list(field_list)
      @field_list = Hash[field_list.map do |type, fields|
        fields = fields.split(',') if fields.is_a? String
        [type, fields]
      end]

      self
    end

    def field_queries(field_queries)
      @field_queries = field_queries

      self
    end

    def field_config(field_config)
      @field_config = field_config

      self
    end

    def filter_by(field_list)
      @search_query.filter_by.merge! field_list

      self
    end

    def filter_or(field_list)
      @search_query.filter_or.merge! field_list

      self
    end

    def filter_by_delimiter(string)
      @search_query.filter_by_delimiter = string

      self
    end

    def filter_or_delimiter(string)
      @search_query.filter_or_delimiter = string

      self
    end

    def sort_by(field_list)
      @search_query.sort_by.merge! field_list

      self
    end

    def match_any
      @search_query.match_any = true

      self
    end

    def page(page)
      @search_query.page = page.to_i

      self
    end

    def per_page(per_page)
      @search_query.per_page = per_page.to_i

      self
    end

    # Performs the search
    def search
      @raw_results = @search_query.customizable.class.__elasticsearch__.search(query_hash)

      self
    end

    # Returns the results after search has been performed
    # Call search() first!
    def results
      return nil if @raw_results.nil?

      begin
        @raw_results.count

        @raw_results.results
      rescue Faraday::ConnectionFailed => e
        @search_query.customizable.class.handle_search_connection_error(e)
      end
    end

    # Fetches the ActiveRecord database records
    # Does not perform search, so call search() first!
    def records
      return nil if @raw_results.nil?

      begin
        @raw_results.count

        @raw_results.records
      rescue Faraday::ConnectionFailed => e
        @search_query.customizable.class.handle_search_connection_error(e)
      end
    end

    # Performs the search and returns the count of results
    def count
      search.results.count
    rescue Faraday::ConnectionFailed => e
      @search_query.customizable.class.handle_search_connection_error(e)
    end

    # Use at own risk, no Connection Exception handling here
    def raw
      @raw_results
    end

    # Return the hash that is conform to elasticsearch's DSL
    # Does not perform a search query!
    def query_hash
      process_field_list

      @search_query.build
    end

    private

    # Combine field list, field queries and field config into one field list
    def process_field_list
      # apply default fields if no user specified fields are present
      if @field_list.empty?
        field_list = @search_query.default_fields
      else
        field_list = Hash[@field_list.map { |type, list| [type, Hash[list.map { |v| [v.to_sym, {}] }]] }]
      end

      # field list is in the format {fields: {field1: {}, field2: {},...}, custom_fields: {field1: {}, ...}}
      # this mapper has the purpose to populate the hash with additional configuration on a per-field basis
      #
      # type: either :fields or :custom_fields
      # list: a hash of fields with options
      field_list.map do |type, list|
        [
          type,
          Hash[list.map do |slug, options|
            # apply options already present in the field_list (possible default settings)
            field_options = options

            # overwrite query (fulltext search) with field specific setting
            if @field_queries[type] && @field_queries[type][slug]
              field_options[:query] = @field_queries[type][slug]
            end

            # add other configuration such as fuzziness, operator, ...
            if @field_config[type] && @field_config[type][slug]
              field_options.merge! @field_config[type][slug]
            end

            [slug, field_options]
          end]
        ]
      end

      @search_query.field_list = field_list
    end
  end
end
